#include "dll.h"
#include "sock.h"
#include <cstdio>

#define MAX_HTTP 16
#define READSTEP (1024*10)

// type definitions

typedef struct {
	bool active;
	sock *s;
	char *method;
	char *filename;
	char *headers_in;
	unsigned long message_body_in_length;
	char *message_body_in;
	short status;
	short action;
	char *statuscode;
	char *statuscode_text;
	char *headers;
	unsigned long message_body_length;
	char *message_body;
} http;

// global variables

bool initialize = true;
bool valid = false;

char errorstring[1] = {'\0'};
char* returnstring = NULL;

http httparray[16];

// #################### speciale functies ####################

bool init() {
	short i;
	if(initialize) {
		if(!sock_init()) {
			valid = false;
		} else {
			for(i=0;i<MAX_HTTP;++i) {
				httparray[i].active = false;
			}
			valid = true;
		}
		initialize = false;
	}
	return valid;
}

bool http_activate(http* h) {
	h->active = true;
	h->s = NULL;
	h->method = NULL;
	h->filename = NULL;
	h->headers_in = NULL;
	h->message_body_in_length = 0;
	h->message_body_in = NULL;
	h->status = 0;
	h->action = 0;
	h->statuscode = NULL;
	h->statuscode_text = NULL;
	h->headers = NULL;
	h->message_body_length = 0;
	h->message_body = NULL;
	return true;
}

bool http_deactivate(http* h) {
	h->active = false;
	sock_destroy(h->s);
	free(h->method);
	free(h->filename);
	free(h->headers_in);
	free(h->message_body_in);
	free(h->statuscode);
	free(h->statuscode_text);
	free(h->headers);
	free(h->message_body);
	return true;
}

// #################### http functies ####################

export double http_initerror() {
	if(!init()) return 1;
	return 0;
}

export double http_startthread() {
	if(!init()) return -1;
	if(sockthread_start()) return 1;
	else return 0;
}

export double http_stopthread() {
	if(!init()) return -1;
	if(sockthread_stop()) return 1;
	else return 0;
}

export double http_handlesockets() {
	if(!init()) return -1;
	sockthread_lock();
	sock_handler();
	sockthread_unlock();
	return 1;
}

export double http_create() {
	if(!init()) return -1;
	short id;
	for(id=0;id<MAX_HTTP;++id) {
		if(!httparray[id].active) {
			http_activate(&httparray[id]);
			return id;
		}
	}
	return -1;
}

export double http_destroy(double http_id) {
	if(!init()) return -1;
	short id;
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			http_deactivate(&httparray[id]);
			return 1;
		}
	}
	return -1;
}

export double http_set_method(double http_id, char* str_method) {
	if(!init()) return -1;
	short id;
	unsigned long len;
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			len = strlen(str_method);
			if(!safe_malloc(&httparray[id].method,len+1)) return 0;
			memcpy(httparray[id].method,str_method,len);
			httparray[id].method[len] = '\0';
			return 1;
		}
	}
	return -1;
}

export double http_set_filename(double http_id, char* str_filename) {
	if(!init()) return -1;
	short id;
	unsigned long len;
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			len = strlen(str_filename);
			if(!safe_malloc(&httparray[id].filename,len+1)) return 0;
			memcpy(httparray[id].filename,str_filename,len);
			httparray[id].filename[len] = '\0';
			return 1;
		}
	}
	return -1;
}

export double http_set_headers_in(double http_id, char* str_headers_in) {
	if(!init()) return -1;
	short id;
	unsigned long i,pos;
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			safe_free(&httparray[id].headers_in);
			for(i=0;str_headers_in[i]!='\0';++i) {
				for(pos=i;str_headers_in[i]!='\r' && str_headers_in[i]!='\n' && str_headers_in[i]!='\0';++i);
				cpytostr(&httparray[id].headers_in,str_headers_in+pos,i-pos);
				cpytostr(&httparray[id].headers_in,"\r\n",2);
				if(str_headers_in[i]=='\0') break;
				if(str_headers_in[i]=='\r' && str_headers_in[i+1]=='\n') ++i;
			}
			return 1;
		}
	}
	return -1;
}

export double http_set_message_body_in(double http_id, char* str_message_body_in) {
	if(!init()) return -1;
	short id;
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			httparray[id].message_body_in_length = strlen(str_message_body_in);
			if(!safe_malloc(&httparray[id].message_body_in,httparray[id].message_body_in_length)) {
				httparray[id].message_body_in_length = 0;
				return 0;
			}
			memcpy(httparray[id].message_body_in,str_message_body_in,httparray[id].message_body_in_length);
			return 1;
		}
	}
	return -1;
}

export double http_load_message_body_in(double http_id, char* filename) {
	if(!init()) return -1;
	short id;
	unsigned long i;
	FILE *f;
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			safe_free(&httparray[id].message_body_in);
			httparray[id].message_body_in_length = 0;
			f = fopen(filename,"rb");
			if(f==NULL) return 0;
			while(true) {
				if(!safe_realloc(&httparray[id].message_body_in,httparray[id].message_body_in_length+READSTEP)) {
					safe_free(&httparray[id].message_body_in);
					fclose(f);
					return 0;
				}
				i = fread(httparray[id].message_body_in+httparray[id].message_body_in_length,1,READSTEP,f);
				if(i==0) break;
				httparray[id].message_body_in_length += i;
			}
			fclose(f);
			return 1;
		}
	}
	return -1;
}

// ##### ##### ##### ##### ##### ##### #####

export double http_connect(double http_id, char* hostname, char* port) {
	if(!init()) return -1;
	short id;
	bool a;
	char str[20];
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			if(httparray[id].status!=1) {
				
				if(httparray[id].method==NULL) return 0;
				if(httparray[id].filename==NULL) return 0;
				
				httparray[id].status = 0;
				httparray[id].action = 0;
				safe_free(&httparray[id].statuscode);
				safe_free(&httparray[id].statuscode_text);
				safe_free(&httparray[id].headers);
				httparray[id].message_body_length = 0;
				safe_free(&httparray[id].message_body);
				
				sockthread_lock();
				httparray[id].s = sock_create();
				if(httparray[id].s==NULL) {
					sockthread_unlock();
					return 0;
				}
				sock_lock(httparray[id].s);
				sockthread_unlock();
				
				if(!sock_connect(httparray[id].s,hostname,port)) {
					sockthread_lock();
					sock_destroy(httparray[id].s);
					httparray[id].s = NULL;
					sockthread_unlock();
					return 0;
				}
				
				sockthread_lock();
				sock_unlock(httparray[id].s);
				a = true;
				a = a && sock_write_str(httparray[id].s,httparray[id].method);
				a = a && sock_write_str(httparray[id].s," ");
				a = a && sock_write_str(httparray[id].s,httparray[id].filename);
				a = a && sock_write_str(httparray[id].s," HTTP/1.0\r\n");
				a = a && sock_write_str(httparray[id].s,"Host: ");
				a = a && sock_write_str(httparray[id].s,hostname);
				a = a && sock_write_str(httparray[id].s,"\r\n");
				a = a && sock_write_str(httparray[id].s,"Connection: close\r\n");
				if(httparray[id].headers_in!=NULL) {
					a = a && sock_write_str(httparray[id].s,httparray[id].headers_in);
				}
				a = a && sock_write_str(httparray[id].s,"\r\n");
				if(httparray[id].message_body_in!=NULL) {
					a = a && sock_write(httparray[id].s,httparray[id].message_body_in,httparray[id].message_body_in_length);
				}
				if(!a) {
					sock_destroy(httparray[id].s);
					httparray[id].s = NULL;
					sockthread_unlock();
					return 0;
				}
				sockthread_unlock();
				
				httparray[id].status = 1;
				return 1;
			}
			return 0;
		}
	}
	return -1;
}

export double http_disconnect(double http_id) {
	if(!init()) return -1;
	short id;
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			if(httparray[id].status==1) {
				sockthread_lock();
				sock_destroy(httparray[id].s);
				httparray[id].s = NULL;
				sockthread_unlock();
				httparray[id].status = 3;
				return 1;
			}
			return 0;
		}
	}
	return -1;
}

export double http_update(double http_id) {
	if(!init()) return -1;
	short id;
	char* str;
	short a;
	unsigned long str_len,pos,len;
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			if(httparray[id].status==1) {
				sockthread_lock();
				while(true) {
					if(httparray[id].action==0) { // wacht op statuscode
						if(sock_read_line(httparray[id].s,&str,&str_len)) {
							pos = 0; len = 0;
							next_word(str,str_len,&pos,&len); // HTTP type
							next_word(str,str_len,&pos,&len); // HTTP statuscode
							cpytostr(&httparray[id].statuscode,str+pos,len);
							next_word(str,str_len,&pos,&len); // HTTP statuscode string
							len = str_len-pos;
							cpytostr(&httparray[id].statuscode_text,str+pos,len);
							httparray[id].action = 1;
							continue;
						}
					}
					if(httparray[id].action==1) { // wacht op headers
						if(sock_read_line(httparray[id].s,&str,&str_len)) {
							if(str_len==0) {
								httparray[id].action = 2;
							} else {
								cpytostr(&httparray[id].headers,str,str_len);
								cpytostr(&httparray[id].headers,"\r\n",2);
							}
							continue;
						}
					}
					if(httparray[id].action==2) {
						sock_read_left(httparray[id].s,&str,&str_len);
						if(!safe_realloc(&httparray[id].message_body,httparray[id].message_body_length+str_len+1)) {
							sock_destroy(httparray[id].s);
							httparray[id].s = NULL;
							httparray[id].status = 3;
							break;
						}
						memcpy(httparray[id].message_body+httparray[id].message_body_length,str,str_len);
						httparray[id].message_body[httparray[id].message_body_length+str_len] = '\0';
						httparray[id].message_body_length += str_len;
					}
					sock_cleanup(httparray[id].s);
					a = sock_status(httparray[id].s);
					if(a==0) {
						sock_destroy(httparray[id].s);
						httparray[id].s = NULL;
						httparray[id].status = 2;
					}
					if(a==3) {
						sock_destroy(httparray[id].s);
						httparray[id].s = NULL;
						httparray[id].status = 3;
					}
					break;
				}
				sockthread_unlock();
				return 1;
			}
			return 0;
		}
	}
	return -1;
}

// ##### ##### ##### ##### ##### ##### #####

export double http_get_conn_status(double http_id) {
	if(!init()) return -1;
	short id;
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			return httparray[id].status;
		}
	}
	return -1;
}

export double http_get_action(double http_id) {
	if(!init()) return -1;
	short id;
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			return httparray[id].action;
		}
	}
	return -1;
}

export char* http_get_statuscode(double http_id) {
	if(!init()) return errorstring;
	short id;
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			if(httparray[id].statuscode==NULL) return errorstring;
			else return httparray[id].statuscode;
		}
	}
	return errorstring;
}

export char* http_get_statuscode_text(double http_id) {
	if(!init()) return errorstring;
	short id;
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			if(httparray[id].statuscode_text==NULL) return errorstring;
			else return httparray[id].statuscode_text;
		}
	}
	return errorstring;
}

export char* http_get_headers(double http_id) {
	if(!init()) return errorstring;
	short id;
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			if(httparray[id].headers==NULL) return errorstring;
			else return httparray[id].headers;
		}
	}
	return errorstring;
}

export char* http_get_header_value(double http_id, char* headername) {
	if(!init()) return errorstring;
	short id;
	bool found;
	unsigned long i,start,pos,len,str_len;
	char* str;
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			str_len = strlen(headername);
			str = httparray[id].headers;
			if(str==NULL) {
				return errorstring;
			} else {
				found = false;
				for(i=0;str[i]!='\0';++i) {
					for(start=i;str[i]!='\r' && str[i]!='\n' && str[i]!='\0';++i);
					pos = start;
					len = 0;
					next_word(str,i,&pos,&len);
					if(pos==start) {
						if(found) {
							return returnstring;
						} else {
							if(len!=0 && len-1==str_len) {
								if(str[pos+len-1]==':' && memicmp(str+pos,headername,str_len)==0) {
									found = true;
									if(!safe_malloc(&returnstring,1)) return errorstring;
									returnstring[0] = '\0';
									if(next_word(str,i,&pos,&len)) {
										cpytostr(&returnstring,str+pos,len);
										while(next_word(str,i,&pos,&len)) {
											cpytostr(&returnstring," ",1);
											cpytostr(&returnstring,str+pos,len);
										}
									}
								}
							}
						}
					} else {
						if(found) {
							cpytostr(&returnstring," ",1);
							cpytostr(&returnstring,str+pos,len);
							while(next_word(str,i,&pos,&len)) {
								cpytostr(&returnstring," ",1);
								cpytostr(&returnstring,str+pos,len);
							}
						}
					}
					if(str[i]=='\0') break;
					if(str[i]=='\r' && str[i+1]=='\n') ++i;
				}
				if(found) return returnstring;
				return errorstring;
			}
		}
	}
	return errorstring;
}

export double http_get_message_body_length(double http_id) {
	if(!init()) return -1;
	short id;
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			return httparray[id].message_body_length;
		}
	}
	return -1;
}

export char* http_get_message_body(double http_id) {
	if(!init()) return errorstring;
	short id;
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			if(httparray[id].message_body==NULL) return errorstring;
			else return httparray[id].message_body;
		}
	}
	return errorstring;
}

export double http_save_message_body(double http_id, char* filename) {
	if(!init()) return -1;
	short id;
	FILE *f;
	id = (short) http_id;
	if(0<=id && id<16) {
		if(httparray[id].active) {
			f = fopen(filename,"wb");
			if(f==NULL) return 0;
			if(httparray[id].message_body!=NULL) {
				fwrite(httparray[id].message_body,1,httparray[id].message_body_length,f);
			}
			if(fclose(f)==EOF) return 0;
			return 1;
		}
	}
	return -1;
}

export char* http_url_get_hostname(char* url) {
	char *pos1,*pos2,*pos3;
	pos1 = strstr(url,"://");
	if(pos1==NULL) pos1 = url;
	else pos1 += 3;
	pos2 = strstr(pos1,"/");
	if(pos2==NULL) pos2 = url+strlen(url);
	pos3 = strstr(pos1,":");
	if(pos3==NULL || pos3>pos2) pos3 = pos2;
	if(!safe_malloc(&returnstring,pos3-pos1+1)) return errorstring;
	memcpy(returnstring,pos1,pos3-pos1);
	returnstring[pos3-pos1] = '\0';
	return returnstring;
}

export char* http_url_get_port(char* url) {
	char *pos1,*pos2,*pos3;
	pos1 = strstr(url,"://");
	if(pos1==NULL) pos1 = url;
	else pos1 += 3;
	pos2 = strstr(pos1,"/");
	if(pos2==NULL) pos2 = url+strlen(url);
	pos3 = strstr(pos1,":");
	if(pos3==NULL || pos3>pos2) pos3 = pos2;
	if(pos3==pos2) {
		if(!safe_malloc(&returnstring,3)) return errorstring;
		memcpy(returnstring,"80",2);
		returnstring[2] = '\0';
	} else {
		if(!safe_malloc(&returnstring,pos2-pos3)) return errorstring;
		memcpy(returnstring,pos3+1,pos2-pos3-1);
		returnstring[pos2-pos3-1] = '\0';
	}
	return returnstring;
}

export char* http_url_get_filename(char* url) {
	char *pos1,*pos2,*pos3;
	pos1 = strstr(url,"://");
	if(pos1==NULL) pos1 = url;
	else pos1 += 3;
	pos3 = strstr(url,"#");
	if(pos3==NULL) pos3 = url+strlen(url); // positie van de NULL byte of #
	pos2 = strstr(pos1,"/");
	if(pos2==NULL) pos2 = pos3;
	else pos2 += 1;
	if(!safe_malloc(&returnstring,pos3-pos2+2)) return errorstring;
	returnstring[0] = '/';
	memcpy(returnstring+1,pos2,pos3-pos2);
	returnstring[pos3-pos2+1] = '\0';
	return returnstring;
}

#define url_ishex(c) (('0'<=c && c<='9') || ('a'<=c && c<='f') || ('A'<=c && c<='F'))
#define url_isbasic(c) (('0'<=c && c<='9') || ('a'<=c && c<='z') || ('A'<=c && c<='Z') || c=='-' || c=='_')
#define url_isspecial(c) (c=='%' || c=='$' || c=='.' || c=='+' || c=='!' || c=='*' || c=='\'' || c=='(' || c==')' || c==',' || c==';' || c=='/' || c=='?' || c==':' || c=='@' || c=='=' || c=='&' || c=='#')

inline char fromhex(char c) {
	if('0'<=c && c<='9') return c-'0';
	if('a'<=c && c<='f') return c-'a'+10;
	if('A'<=c && c<='F') return c-'A'+10;
	return 0;
}

export char* http_urlencode(char* input, double keepspecialchars) {
	short m;
	char c;
	char hex[16] = {'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};
	unsigned long i,j;
	m = (short) keepspecialchars;
	j = 0;
	for(i=0;c=input[i],c!='\0';++i) {
		if(url_isbasic(c)) {
			j += 1;
		} else if(m>0 && url_isspecial(c)) {
			j += 1;
		} else {
			j += 3;
		}
	}
	if(!safe_malloc(&returnstring,j+1)) return errorstring;
	j = 0;
	for(i=0;c=input[i],c!='\0';++i) {
		if(url_isbasic(c)) {
			returnstring[j++] = c;
		} else if(m>0 && url_isspecial(c)) {
			returnstring[j++] = c;
		} else {
			returnstring[j++] = '%';
			returnstring[j++] = hex[(c>>4)&0xf];
			returnstring[j++] = hex[c&0xf];
		}
	}
	returnstring[j] = '\0';
	return returnstring;
}

export char* http_urldecode(char* input) {
	char c,d,e;
	unsigned long i,j,len;
	len = strlen(input);
	j = 0;
	for(i=0;i<len;++i) {
		c = input[i];
		if(i+2<len) {
			d = input[i+1]; e = input[i+2];
			if(c=='%' && url_ishex(d) && url_ishex(e)) {
			i += 2;
			++j;
			continue;
			}
		}
		++j;
	}
	if(!safe_malloc(&returnstring,j+1)) return errorstring;
	j = 0;
	for(i=0;i<len;++i) {
		c = input[i];
		if(i+2<len) {
			d = input[i+1]; e = input[i+2];
			if(c=='%' && url_ishex(d) && url_ishex(e)) {
			i += 2;
			returnstring[j++] = (fromhex(d)<<4)|fromhex(e);
			continue;
			}
		}
		if(c=='+') c = ' ';
		returnstring[j++] = c;
	}
	returnstring[j] = '\0';
	return returnstring;
}

