#include "sock.h"

// global variables

WSADATA wsaData;

Mutex sockthread_mutex = NULL;

bool sockthread_active = false;
HANDLE sockthread; 
DWORD sockthread_id;

sock sockarray[16];

// basic functions

#define Mutex_init(mut) {\
	mut = CreateMutex(NULL,FALSE,NULL);\
}
#define Mutex_deinit(mut) {\
	CloseHandle(mut);\
	mut = NULL;\
}
#define Mutex_lock(mut) {\
	WaitForSingleObject(mut,INFINITE);\
}
#define Mutex_unlock(mut) {\
	ReleaseMutex(mut);\
}

bool safe_malloc(char** ptr, unsigned long size) {
	free(*ptr);
	if(size==0) {
		return true;
	} else {
		*ptr = (char*) malloc(size);
		if(*ptr!=NULL) {
			return true;
		} else {
			return false;
		}
	}
}

bool safe_realloc(char** ptr, unsigned long size) {
	void* temp;
	temp = realloc(*ptr,size);
	if(temp!=NULL || size==0) {
		*ptr = (char*) temp;
		return true;
	} else {
		return false;
	}
}

bool safe_free(char** ptr) {
	free(*ptr);
	*ptr = NULL;
	return true;
}

bool next_word(const char* str, unsigned long str_len, unsigned long* pos, unsigned long* len) {
	bool isspace;
	bool inword = false;
	unsigned long i;
	for(i=*pos+*len;i<str_len;++i) {
		isspace = (str[i]==' ' || str[i]=='\t');
		if(!inword && !isspace) {
			*pos = i;
			inword = true;
		}
		if(inword && isspace) {
			*len = i-*pos;
			return true;
		}
	}
	if(inword) {
		*len = str_len-*pos;
		return true;
	} else {
		*pos = str_len;
		*len = 0;
		return false;
	}
}

bool cpytostr(char** str, const char* source, unsigned long len) {
	unsigned long i,str_len;
	for(i=0;i<len;++i) {
		if(source[i]=='\0') {
			len = i;
			break;
		}
	}
	if(*str==NULL) str_len = 0;
	else str_len = strlen(*str);
	if(!safe_realloc(str,str_len+len+1)) return false;
	memcpy(*str+str_len,source,len);
	(*str)[str_len+len] = '\0';
	return true;
}

// socket functions

bool sock_activate(sock* s) {
	s->active = true;
	s->locked = false;
	s->status = 0;
	s->readbuffer = NULL;
	s->writebuffer = NULL;
	s->readpos = 0;
	s->readlen = 0;
	s->writepos = 0;
	s->writelen = 0;
	return true;
}

bool sock_deactivate(sock* s) {
	s->active = false;
	if(s->status==1 || s->status==2) {
		closesocket(s->socket);
	}
	free(s->readbuffer); s->readbuffer = NULL;
	free(s->writebuffer); s->writebuffer = NULL;
	return true;
}

bool sock_lock(sock* s) {
	if(s->active) {
		s->locked = true;
	}
}

bool sock_unlock(sock* s) {
	if(s->active) {
		s->locked = false;
	}
}

bool sock_connect(sock* s, const char* hostname, const char* port) {
	int ii;
	u_long on = 1;
	struct addrinfo hints;
	if(s->active) {
		if(s->status==0 || s->status==3) {
			s->status = 0;
			free(s->readbuffer); s->readbuffer = NULL;
			s->readpos = 0;
			s->readlen = 0;
			free(s->writebuffer); s->writebuffer = NULL;
			s->writepos = 0;
			s->writelen = 0;
			memset(&hints,0,sizeof(hints));
		    hints.ai_family = AF_UNSPEC;
		    hints.ai_socktype = SOCK_STREAM;
		    hints.ai_protocol = IPPROTO_TCP;
		    s->addr_result = NULL;
		    ii = getaddrinfo(hostname,port,&hints,&(s->addr_result));
		    if(ii!=0) {
		        return false;
		    }
		    for(s->addr_ptr = s->addr_result;s->addr_ptr!=NULL;s->addr_ptr=s->addr_ptr->ai_next) {
				s->socket = socket(s->addr_ptr->ai_family,s->addr_ptr->ai_socktype,s->addr_ptr->ai_protocol);
				if(s->socket==INVALID_SOCKET) {
				    continue;
				}
				ioctlsocket(s->socket,FIONBIO,&on);
				connect(s->socket,s->addr_ptr->ai_addr,s->addr_ptr->ai_addrlen);
				s->status = 1;
				return true;
			}
			freeaddrinfo(s->addr_result);
		    return false;
		}
	}
	return false;
}

bool sock_disconnect(sock* s) {
	if(s->active) {
		if(s->status==1 || s->status==2) {
			closesocket(s->socket);
			s->status = 0;
			return true;
		}
	}
	return false;
}

bool sock_write(sock* s, const char* data, unsigned long len) {
	if(s->active) {
		if(!safe_realloc(&(s->writebuffer),(s->writelen)+len)) {
			return false;
		}
		memcpy((s->writebuffer)+(s->writelen),data,len);
		s->writelen += len;
		return true;
	}
	return false;
}

bool sock_write_str(sock* s, const char* str) {
	return sock_write(s,str,strlen(str));
}

bool sock_read_line(sock* s, char** data, unsigned long* len) {
	unsigned long pos;
	if(s->active) {
		for(pos=(s->readpos);pos<(s->readlen);++pos) {
			if((s->readbuffer[pos])=='\r') {
				*data = (s->readbuffer)+(s->readpos);
				*len = pos-(s->readpos);
				if(++pos<(s->readlen)) {
					if((s->readbuffer[pos])=='\n') ++pos;
				}
				s->readpos = pos;
				return true;
			}
			if((s->readbuffer[pos])=='\n') {
				*data = (s->readbuffer)+(s->readpos);
				*len = pos-(s->readpos);
				++pos;
				s->readpos = pos;
				return true;
			}
		}
	}
	return false;
}

bool sock_read_chars(sock* s, char** data, unsigned long len) {
	unsigned long pos;
	if(s->active) {
		if((s->readpos)+len<=s->readlen) {
			*data = (s->readbuffer)+(s->readpos);
			s->readpos += len;
			return true;
		} else {
			return false;
		}
	}
	return false;
}

bool sock_read_left(sock* s, char** data, unsigned long* len) {
	if(s->active) {
		*data = (s->readbuffer)+(s->readpos);
		*len = (s->readlen)-(s->readpos);
		s->readpos = s->readlen;
		return true;
	}
	return false;
}

short sock_status(sock* s) {
	short out = -1;
	if(s->active) {
		out = s->status;
	}
	return out;
}

bool sock_clear(sock* s) {
	if(s->active) {
		free(s->readbuffer); s->readbuffer = NULL;
		s->readpos = 0;
		s->readlen = 0;
		free(s->writebuffer); s->writebuffer = NULL;
		s->writepos = 0;
		s->writelen = 0;
		return true;
	}
	return false;
}

bool sock_cleanup(sock* s) {
	if(s->active) {
		if((s->readpos)>0) {
			memmove(s->readbuffer,(s->readbuffer)+(s->readpos),(s->readlen)-(s->readpos));
			s->readlen -= s->readpos;
			s->readpos = 0;
			safe_realloc(&(s->readbuffer),s->readlen);
		}
		if((s->writepos)>0) {
			memmove(s->writebuffer,(s->writebuffer)+(s->writepos),(s->writelen)-(s->writepos));
			s->writelen -= s->writepos;
			s->writepos = 0;
			safe_realloc(&(s->writebuffer),s->writelen);
		}
		return true;
	}
	return false;
}

sock* sock_create() {
	short i;
	for(i=0;i<MAX_SOCK;++i) {
		if(!sockarray[i].active) {
			sock_activate(&sockarray[i]);
			return &sockarray[i];
		}
	}
	return NULL;
}

void sock_destroy(sock* s) {
	if(s!=NULL) {
		if(s->active) {
			sock_deactivate(s);
		}
	}
}

void sock_handler() {
	
	bool a;
	short i,c;
	int ii;
	u_long on = 1;
	fd_set fd_r;
	fd_set fd_w;
	fd_set fd_e;
	timeval tv;
	
	tv.tv_sec = 0;
	tv.tv_usec = 0;
	
	a = false;
	FD_ZERO(&fd_r);
	FD_ZERO(&fd_w);
	FD_ZERO(&fd_e);
	
	for(i=0;i<MAX_SOCK;++i) {
		if(sockarray[i].active && !sockarray[i].locked) {
			if(sockarray[i].status==1) {
				a = true;
				FD_SET(sockarray[i].socket,&fd_w);
				FD_SET(sockarray[i].socket,&fd_e);
			} else if(sockarray[i].status==2) {
				a = true;
				FD_SET(sockarray[i].socket,&fd_r);
				FD_SET(sockarray[i].socket,&fd_w);
				FD_SET(sockarray[i].socket,&fd_e);
			}
		}
	}
	if(a) {
		ii = select(0,&fd_r,&fd_w,&fd_e,&tv);
		if(ii!=SOCKET_ERROR) {
			for(i=0;i<MAX_SOCK;++i) {
				if(sockarray[i].active && !sockarray[i].locked) {
					if(sockarray[i].status==1) {
						if(FD_ISSET(sockarray[i].socket,&fd_w)) {
							freeaddrinfo(sockarray[i].addr_result);
							sockarray[i].status = 2;
						}
						if(FD_ISSET(sockarray[i].socket,&fd_e)) {
							closesocket(sockarray[i].socket);
							a = false;
						    for(sockarray[i].addr_ptr=sockarray[i].addr_ptr->ai_next;sockarray[i].addr_ptr!=NULL;sockarray[i].addr_ptr=sockarray[i].addr_ptr->ai_next) {
								sockarray[i].socket = socket(sockarray[i].addr_ptr->ai_family,sockarray[i].addr_ptr->ai_socktype,sockarray[i].addr_ptr->ai_protocol);
								if(sockarray[i].socket==INVALID_SOCKET) {
								    continue;
								}
								ioctlsocket(sockarray[i].socket,FIONBIO,&on);
								connect(sockarray[i].socket,sockarray[i].addr_ptr->ai_addr,sockarray[i].addr_ptr->ai_addrlen);
								a = true;
								break;
							}
							if(!a) {
								freeaddrinfo(sockarray[i].addr_result);
								sockarray[i].status = 3;
							}
						}
					} else if(sockarray[i].status==2) {
						if(FD_ISSET(sockarray[i].socket,&fd_r)) {
							if(safe_realloc(&sockarray[i].readbuffer,(sockarray[i].readlen)+RECVSTEP)) {
								ii = recv(sockarray[i].socket,(sockarray[i].readbuffer)+(sockarray[i].readlen),RECVSTEP,0);
								if(ii>0) {
									sockarray[i].readlen += ii;
								} else if(ii==0) {
									closesocket(sockarray[i].socket);
									sockarray[i].status = 0;
								} else {
									closesocket(sockarray[i].socket);
									sockarray[i].status = 3;
								}
							}
						}
						if(FD_ISSET(sockarray[i].socket,&fd_w)) {
							if((sockarray[i].writelen)-(sockarray[i].writepos)>0) {
								ii = send(sockarray[i].socket,(sockarray[i].writebuffer)+(sockarray[i].writepos),(sockarray[i].writelen)-(sockarray[i].writepos),0);
								if(ii>0) {
									sockarray[i].writepos += ii;
								}
							}
						}
					}
				}
			}
		}
	}
	
}

void sockthread_lock() {
	if(sockthread_mutex!=NULL) {
		Mutex_lock(sockthread_mutex);
	}
}

void sockthread_unlock() {
	if(sockthread_mutex!=NULL) {
		Mutex_unlock(sockthread_mutex);
	}
}

bool sock_init() {
	short i;
	if(WSAStartup(MAKEWORD(2,0),&wsaData)!=0) {
		return false;
	} else if(LOBYTE(wsaData.wVersion)!=2 || HIBYTE(wsaData.wVersion)!=0) {
	    WSACleanup();
		return false;
	} else {
		for(i=0;i<MAX_SOCK;++i) {
			sockarray[i].active = false;
		}
		return true;
	}
}

DWORD sockthread_main() {
	
	bool a;
	
	while(true) {
		
		Mutex_lock(sockthread_mutex);
		sock_handler();
		a = sockthread_active;
		Mutex_unlock(sockthread_mutex);
		
		if(!a) break;
		Sleep(10);
		
	}
	
	return 0;
	
}

bool sockthread_start() {
	if(sockthread==NULL) {
		Mutex_init(sockthread_mutex);
		if(sockthread_mutex==NULL) {
			return false;
		}
		Mutex_lock(sockthread_mutex);
		sockthread_active = true;
		Mutex_unlock(sockthread_mutex);
		sockthread = CreateThread( 
			NULL,                                            // default security attributes
			0,                                               // use default stack size
			(LPTHREAD_START_ROUTINE) sockthread_main,        // thread function name
			NULL,                                            // argument to thread function
			0,                                               // use default creation flags
			&sockthread_id);                                 // returns the thread identifier
		if(sockthread==NULL) return false;
		return true;
	}
	return true;
}

bool sockthread_stop() {
	if(sockthread!=NULL) {
		Mutex_lock(sockthread_mutex);
		sockthread_active = false;
		Mutex_unlock(sockthread_mutex);
		WaitForSingleObject(sockthread,INFINITE);
		sockthread = NULL;
		Mutex_deinit(sockthread_mutex);
	}
	return true;
}

