#include <ws2tcpip.h>

#define MAX_SOCK 16
#define RECVSTEP (1024*10)

// type definitions

typedef HANDLE Mutex;

typedef struct {
	bool active;
	bool locked;
	short status;
	SOCKET socket;
	unsigned long mps;
	struct addrinfo *addr_result;
	struct addrinfo *addr_ptr;
	char *readbuffer,*writebuffer;
	unsigned long readpos,readlen,writepos,writelen;
} sock;

// basic functions

bool safe_malloc(char** ptr, unsigned long size);
bool safe_realloc(char** ptr, unsigned long size);
bool safe_free(char** ptr);

bool next_word(const char* str, unsigned long str_len, unsigned long* pos, unsigned long* len);
bool cpytostr(char** str, const char* source, unsigned long len);

// socket functions

bool sock_lock(sock* s);
bool sock_unlock(sock* s);

bool sock_connect(sock* s, const char* hostname, const char* port);
bool sock_disconnect(sock* s);

bool sock_write(sock* s, const char* data, unsigned long len);
bool sock_write_str(sock* s, const char* str);
bool sock_read_line(sock* s, char** data, unsigned long* len);
bool sock_read_chars(sock* s, char** data, unsigned long len);
bool sock_read_left(sock* s, char** data, unsigned long* len);

short sock_status(sock* s);

bool sock_clear(sock* s);
bool sock_cleanup(sock* s);

sock* sock_create();
void sock_destroy(sock* s);

void sock_handler();
void sockthread_lock();
void sockthread_unlock();

bool sock_init();
bool sockthread_start();
bool sockthread_stop();

